/*****************************************************************************/
/* EHandler.cpp                               Copyright Ladislav Zezula 2001 */
/*---------------------------------------------------------------------------*/
/* A way how to comfortably handle the exceptions with informations about    */
/* the place of an exception. This program works with dbghelp.dll.           */
/*---------------------------------------------------------------------------*/
/*   Date    Ver   Who  Comment                                              */
/* --------  ----  ---  -----------------------------------------------------*/
/* 21.12.01  1.00  Lad  The first version of DebugInfo.cpp                   */
/* 21.12.01  1.01  Lad  Added support of Unicode version                     */
/* 14.01.02  1.02  Lad  Better info when problem with getting debug info     */
/* 15.01.02  1.02  Lad  Possibility to re-raise exception into the debugger  */
/* 03.04.02  1.02  Lad  Renamed to EHandler.cpp (more exact name)            */
/* 03.04.02  1.02  Lad  Support for compiling as part of MFC dll             */
/*****************************************************************************/

#ifdef _USRDLL                          // If compiling as MFC DLL (Including Windows.h
#include "afxwin.h"                     // when part of a MFC DLL makes problem
#else                                   // at SECOND linking (works after making a project,
#include <windows.h>                    // at second link linker errors)
#include <stdio.h>
#endif 

#include "EHandler.h"

// MakePtr is a macro that allows you to easily add to values (including
// pointers) together without dealing with C's pointer arithmetic.  It
// essentially treats the last two parameters as DWORDs.  The first
// parameter is used to typecast the result to the appropriate pointer type.
#define MakePtr(cast, ptr, addValue) (cast)((DWORD_PTR)(ptr) + (addValue))

// When _UNICODE defined, it redirects MODULEENTRY32 to MODULEENTRY32W.
// Here we want to use ANSI version only !!!
#if (MODULEENTRY32 == MODULEENTRY32W)
#undef MODULEENTRY32
#endif

// This array is used to convert the exception code to readable 
static TExceptionName ExcNames[] =
{
    {EXCEPTION_ACCESS_VIOLATION,         "EXCEPTION_ACCESS_VIOLATION"},
    {EXCEPTION_ARRAY_BOUNDS_EXCEEDED,    "EXCEPTION_ARRAY_BOUNDS_EXCEEDED"},
    {EXCEPTION_BREAKPOINT,               "EXCEPTION_BREAKPOINT"},
    {EXCEPTION_DATATYPE_MISALIGNMENT,    "EXCEPTION_DATATYPE_MISALIGNMENT"},
    {EXCEPTION_FLT_DENORMAL_OPERAND,     "EXCEPTION_FLT_DENORMAL_OPERAND"},
    {EXCEPTION_FLT_DIVIDE_BY_ZERO,       "EXCEPTION_FLT_DIVIDE_BY_ZERO"},
    {EXCEPTION_FLT_INEXACT_RESULT,       "EXCEPTION_FLT_INEXACT_RESULT"},
    {EXCEPTION_FLT_INVALID_OPERATION,    "EXCEPTION_FLT_INVALID_OPERATION"},
    {EXCEPTION_FLT_OVERFLOW,             "EXCEPTION_FLT_OVERFLOW"},
    {EXCEPTION_FLT_STACK_CHECK,          "EXCEPTION_FLT_STACK_CHECK"},
    {EXCEPTION_FLT_UNDERFLOW,            "EXCEPTION_FLT_UNDERFLOW"},
    {EXCEPTION_ILLEGAL_INSTRUCTION,      "EXCEPTION_ILLEGAL_INSTRUCTION"},
    {EXCEPTION_IN_PAGE_ERROR,            "EXCEPTION_IN_PAGE_ERROR"},
    {EXCEPTION_INT_DIVIDE_BY_ZERO,       "EXCEPTION_INT_DIVIDE_BY_ZERO"},
    {EXCEPTION_INT_OVERFLOW,             "EXCEPTION_INT_OVERFLOW"},
    {EXCEPTION_INVALID_DISPOSITION,      "EXCEPTION_INVALID_DISPOSITION"},
    {EXCEPTION_NONCONTINUABLE_EXCEPTION, "EXCEPTION_NONCONTINUABLE_EXCEPTION"},
    {EXCEPTION_PRIV_INSTRUCTION,         "EXCEPTION_PRIV_INSTRUCTION"},
    {EXCEPTION_SINGLE_STEP,              "EXCEPTION_SINGLE_STEP"},
    {EXCEPTION_STACK_OVERFLOW,           "EXCEPTION_STACK_OVERFLOW"},
    {0xFFFFFFFF, "EXCEPTION_UNKNOWN_EXCEPTION"}
};

// Local pointers from DbgHelp.dll
static SYMINITIALIZE      _SymInitialize      = NULL;
static SYMSETOPTIONS      _SymSetOptions      = NULL;
static SYMSETSEARCHPATH   _SymSetSearchPath   = NULL;
static SYMGETSEARCHPATH   _SymGetSearchPath   = NULL;
static SYMGETMODULEINFO   _SymGetModuleInfo   = NULL;
static SYMGETLINEFROMADDR _SymGetLineFromAddr = NULL;
static SYMCLEANUP         _SymCleanup         = NULL;
static HINSTANCE          hDbgHelp;

typedef BOOL  (WINAPI * BOOLFUNC)();

// This function determines if we are on an NT platform (returns TRUE) or 9x (returns FALSE)
static BOOL IsWinNT()
{
    OSVERSIONINFO vi;

    // Read information about operating system
    vi.dwOSVersionInfoSize = sizeof(OSVERSIONINFO);
    GetVersionEx(&vi);

    return (BOOL)(vi.dwPlatformId == VER_PLATFORM_WIN32_NT);
}

// This function converts the exception code to the printable form.
// If no known exception code, it returns "EXCEPTION_UNKNOWN_EXCEPTION".
static const char * GetExceptionName(DWORD dwCode)
{
    for(int i = 0; ExcNames[i].ExcCode != 0xFFFFFFFF; i++)
    {
        if(dwCode == ExcNames[i].ExcCode)
            return ExcNames[i].ExcName;
    }
    return ExcNames[i].ExcName;
}

// This function tests if the process is debugged.
// The IsDebuggerPresent is not available in Win95, so we need to obtain it dynamically
static BOOL TestDebugger()
{
    HINSTANCE hKernel32 = LoadLibraryA("Kernel32.dll");
    BOOLFUNC  IsDebugger = (BOOLFUNC)GetProcAddress(hKernel32, "IsDebuggerPresent");
    BOOL      result;

    if(IsDebugger == NULL)
        return FALSE;

    result = IsDebugger();
    FreeLibrary(hKernel32);
    return result;
}

// This function loads the DbgHelp.dll and sets all the used pointers
static int LoadDbgHelp()
{
    // Don't load the library again
    if(hDbgHelp != NULL)
        return ERROR_SUCCESS;

    // Load the library
    if((hDbgHelp = LoadLibraryA("DbgHelp.dll")) == NULL)
        return GetLastError();

    // Retrieve pointers to functions
    if((_SymInitialize    = (SYMINITIALIZE)GetProcAddress(hDbgHelp, "SymInitialize")) == NULL ||
       (_SymGetSearchPath = (SYMGETSEARCHPATH)GetProcAddress(hDbgHelp, "SymGetSearchPath")) == NULL ||
       (_SymSetSearchPath = (SYMSETSEARCHPATH)GetProcAddress(hDbgHelp, "SymSetSearchPath")) == NULL ||
       (_SymSetOptions = (SYMSETOPTIONS)GetProcAddress(hDbgHelp, "SymSetOptions")) == NULL ||
       (_SymGetModuleInfo = (SYMGETMODULEINFO)GetProcAddress(hDbgHelp, "SymGetModuleInfo")) == NULL ||
       (_SymGetLineFromAddr = (SYMGETLINEFROMADDR)GetProcAddress(hDbgHelp, "SymGetLineFromAddr")) == NULL ||
       (_SymCleanup = (SYMCLEANUP)GetProcAddress(hDbgHelp, "SymCleanup")) == NULL)
        return GetLastError();

    return ERROR_SUCCESS;
}

static int UnloadDbgHelp()
{
    if(hDbgHelp != NULL)
        FreeLibrary(hDbgHelp);
    hDbgHelp = NULL;
    return ERROR_SUCCESS;
}

// Given a section name, look it up in the section table and return a
// pointer to its IMAGE_SECTION_HEADER
PIMAGE_SECTION_HEADER GetSectionHeader(PSTR name, PIMAGE_NT_HEADERS pNTHeader)
{
	PIMAGE_SECTION_HEADER section = (PIMAGE_SECTION_HEADER)(pNTHeader+1);
	
    // Walk through all sections
    for(unsigned i = 0; i < pNTHeader->FileHeader.NumberOfSections; i++, section++)
	{
        // If the section name agrees, return its pointer
		if(!strnicmp((char *)section->Name, name, IMAGE_SIZEOF_SHORT_NAME))
			return section;
	}
	return NULL;
}

// This function walks the ".rdata" section and searches the CodeView file format.
// This format normally contains the full path name of the symbol file (.pdb).
static int SearchPdbFile(PIMAGE_DOS_HEADER dosHeader, char * szSymbolFile)
{
	PIMAGE_DEBUG_DIRECTORY debugDir;
	PIMAGE_SECTION_HEADER header;
	PIMAGE_NT_HEADERS pNTHeader = MakePtr(PIMAGE_NT_HEADERS, dosHeader, dosHeader->e_lfanew);
	DWORD_PTR base = (DWORD_PTR)dosHeader;
	DWORD offsetInto_rdata;
	DWORD va_debug_dir;                 // Virtual address of the debug directory
    int   cFormats;                     // Number of debug formats in the file

	// First, verify that the e_lfanew field gave us a reasonable
	// pointer, then verify the PE signature.
	if(IsBadReadPtr(pNTHeader, sizeof(IMAGE_NT_HEADERS)) || pNTHeader->Signature != IMAGE_NT_SIGNATURE)
	    return ERROR_CAN_NOT_COMPLETE;

	// Retrieve the virtual address of the debug directory
    va_debug_dir = pNTHeader->OptionalHeader.DataDirectory[IMAGE_DIRECTORY_ENTRY_DEBUG].VirtualAddress;
	if(va_debug_dir == 0)
		return ERROR_CAN_NOT_COMPLETE;

	// If we found a .debug section, and the debug directory is at the
	// beginning of this section, it looks like a Borland file
	header = GetSectionHeader(".debug", pNTHeader);
	if(header != NULL && (header->VirtualAddress == va_debug_dir))
        return ERROR_BAD_FORMAT;

    // Find Microsoft debug section
    if((header = GetSectionHeader(".rdata", pNTHeader)) == NULL)
        return ERROR_CAN_NOT_COMPLETE;

	// See if there's even any debug directories to speak of...
	cFormats = pNTHeader->OptionalHeader.DataDirectory[IMAGE_DIRECTORY_ENTRY_DEBUG].Size / sizeof(IMAGE_DEBUG_DIRECTORY);
	if(cFormats == 0)
		return ERROR_CAN_NOT_COMPLETE;

	offsetInto_rdata = va_debug_dir - header->VirtualAddress;
	debugDir = MakePtr(PIMAGE_DEBUG_DIRECTORY, base, header->PointerToRawData + offsetInto_rdata);

	for(int i = 0; i < cFormats; i++)
	{
        // CodeView format ?
        if(debugDir->Type == 2)
        {
            strcpy(szSymbolFile, (char *)(base + debugDir->PointerToRawData + 16));
            return ERROR_SUCCESS;
        }
    }
    return ERROR_CAN_NOT_COMPLETE;
}

// This function tests the existence of the pdb file. Sometimes,
// the PDB file cannot be found in the directory where was stored by the compiler
// (e.g. on the network drive inn another company), so we have the choice
// to place it in the current directory
static int PdbFileExists(char * szModule, char * szPdbFile)
{
    char szDrive[_MAX_DRIVE], szDir[_MAX_DIR], szFName[_MAX_FNAME];

    // First of all, we will test the file returned by the previous function
    if(GetFileAttributesA(szPdbFile) != 0xFFFFFFFF)
        return ERROR_SUCCESS;

    // If the required PDB file couldn't be found, try the same directory
    // like the module
    _splitpath(szModule, szDrive, szDir, szFName, NULL);
    _makepath(szPdbFile, szDrive, szDir, szFName, ".pdb");
    if(GetFileAttributesA(szPdbFile) != 0xFFFFFFFF)
        return ERROR_SUCCESS;

    // Otherwise try to search in the current directory
    _makepath(szPdbFile, "", ".\\", szFName, ".pdb");
    if(GetFileAttributesA(szPdbFile) != 0xFFFFFFFF)
        return ERROR_SUCCESS;

    return ERROR_FILE_NOT_FOUND;
}

// This function returns the whole name of the PDB file. It looks it within the
// image of EXE file.
static int GetPdbFileName(char * szModule, char * szSymbolFile)
{
	PIMAGE_DOS_HEADER dosHeader;
	HANDLE hFile = CreateFileA(szModule, GENERIC_READ, FILE_SHARE_READ, NULL, OPEN_EXISTING, FILE_ATTRIBUTE_NORMAL, 0);
	HANDLE hMap;
	LPVOID lpFileBase;
    int    result = ERROR_FILE_NOT_FOUND;
	
    if(hFile == INVALID_HANDLE_VALUE)
        return GetLastError();

	if((hMap = CreateFileMapping(hFile, NULL, PAGE_READONLY, 0, 0, NULL)) != NULL)
    {
        if((lpFileBase = MapViewOfFile(hMap, FILE_MAP_READ, 0, 0, 0)) != NULL)
        {
            dosHeader = (PIMAGE_DOS_HEADER)lpFileBase;

	        if(dosHeader->e_magic == IMAGE_DOS_SIGNATURE)
            {
		        // Try to search the name
                SearchPdbFile(dosHeader, szSymbolFile);
                result = PdbFileExists(szModule, szSymbolFile);
            }
            UnmapViewOfFile(lpFileBase);
        }
        CloseHandle(hMap);
    }
    CloseHandle(hFile);

    return result;
}

// This function walks the modules and tests them by address of exception.
// We have to retrieve the Module32First and Module32Next dynamically, because of
// WinNT 4.0. This function always returns a name, even if the module cannot be found.
typedef HANDLE(WINAPI * CREATESNAP)(DWORD, DWORD);
typedef BOOL  (WINAPI * MODULEWALK)(HANDLE, MODULEENTRY32 *);
typedef BOOL  (WINAPI * CLOSESNAP)(HANDLE);
typedef BOOL  (WINAPI * ENUMMODULES)(HANDLE, HMODULE *, DWORD, LPDWORD);
typedef BOOL  (WINAPI * GETMODINFO)(HANDLE, HMODULE, MODULEINFO *, DWORD);

static HMODULE GetModuleFromAddr(PVOID pAddr, char * szFileName)
{
    MODULEENTRY32 me;
    HINSTANCE  hKernel32 = LoadLibraryA("Kernel32.dll");
    CREATESNAP CreateSnapshot = (CREATESNAP)GetProcAddress(hKernel32, "CreateToolhelp32Snapshot");
    MODULEWALK Module32First = (MODULEWALK)GetProcAddress(hKernel32, "Module32First");
    MODULEWALK Module32Next  = (MODULEWALK)GetProcAddress(hKernel32, "Module32Next");
    HMODULE    hModule = NULL;
    PBYTE      pbAddr = (PBYTE)pAddr;
    BOOL       result;

    // Default returned module name
    GetModuleFileNameA(NULL, szFileName, MAX_PATH);

    // Test if we retrieved the function addresses successfully (Win9x, Win2000, WinXP)
    if(CreateSnapshot != NULL && Module32First != NULL && Module32Next != NULL)
    {
        HANDLE  hSnap = CreateSnapshot(TH32CS_SNAPMODULE, 0);

        me.dwSize = sizeof(MODULEENTRY32);
        result = Module32First(hSnap, &me);
        while(result == TRUE)
        {
            if(me.modBaseAddr <= pbAddr && pbAddr <= me.modBaseAddr + me.modBaseSize)
            {
                strcpy(szFileName, me.szExePath);
                hModule = me.hModule;
                break;
            }
            result = Module32Next(hSnap, &me);
        }
        CloseHandle(hSnap);
    }
    else
    {
        // Version for WinNT 4.0
        MODULEINFO  mi;
        HINSTANCE   hPSAPI = LoadLibraryA("PSAPI.dll");
        ENUMMODULES EnumProcessModules = (ENUMMODULES)GetProcAddress(hPSAPI, "EnumProcessModules");
        GETMODINFO  GetModuleInformation = (GETMODINFO)GetProcAddress(hPSAPI, "GetModuleInformation");
        HMODULE   * hModules = NULL;
        HANDLE      hProcess = GetCurrentProcess();
        DWORD       dwModules;
        PBYTE       pbModuleBase;

        // Retrieve the numkber of modules
        EnumProcessModules(hProcess, hModules, 0, &dwModules);
        hModules = new HMODULE[dwModules / sizeof(HMODULE)];
        EnumProcessModules(hProcess, hModules, dwModules, &dwModules);

        // Walk all returned module handles
        for(DWORD i = 0; i < dwModules / sizeof(HMODULE); i++)
        {
            if(GetModuleInformation(hProcess, hModules[i], &mi, sizeof(MODULEINFO)) == FALSE)
                break;
            
            pbModuleBase = (PBYTE)mi.lpBaseOfDll;
            if(pbModuleBase <= pbAddr && pbAddr <= pbModuleBase + mi.SizeOfImage)
            {
                GetModuleFileNameA(hModules[i], szFileName, MAX_PATH);
                hModule = hModules[i];
                return FALSE;
            }
        }
        delete hModules;
        FreeLibrary(hPSAPI);
    }

    // Cleanup and exit.
    FreeLibrary(hKernel32);
    return hModule;
}

// This function converts the address to source file name and line number.
// Returns the Win32 error code. If succeeded (must be a debug version) it returns
// ERROR_SUCCESS. In the "szModuleName" is stored the module name (if not 0 at the begin).
// If this function fails, then the szFileName contains the reason why coudn't
// load the debug info.
int GetSourceNameAndLineNum(PVOID pAddr, char * szFileName, DWORD & dwLine)
{
    IMAGEHLP_LINE il;
    HANDLE hProcess;
    HANDLE hModule;                 // Where the exception occured
    BOOL   bResult;
    char   szSearchDir[MAX_PATH*3]; // Directory for finding the PDB file
    char   szDrive[_MAX_DRIVE+1], szDir[_MAX_DIR+1];
    char   szTemp[MAX_PATH];        // "Self" file name
    DWORD  dwDisplacement;          // Distance from the line
    int    result = ERROR_CAN_NOT_COMPLETE;

    // Under WinNT, we have to use process handle, otherwise process ID
    hProcess = GetCurrentProcess();
    if(IsWinNT() == FALSE)
        hProcess = (HANDLE)(DWORD_PTR)GetCurrentProcessId();

    // Retrieve the information about the module where the exception occurred
    // I tried to use DbgHelp function, but it has problems with new settings of
    // search path. So we will try to find the module before we initialize the
    // DbgHelp.dll. If this function fails, hModule becomes NULL.
    hModule = GetModuleFromAddr(pAddr, szTemp);

    // Retrieve the directory where to find the PDB file. If the name was not
    // found in the module itself, we have to try to find it in the same directory
    // like the module resides.
    if(GetPdbFileName(szTemp, szSearchDir) != ERROR_SUCCESS)
    {
        sprintf(szFileName, "PDB file with debug info was not found");
        return ERROR_FILE_NOT_FOUND;
    }
    _splitpath(szSearchDir, szDrive, szDir, NULL, NULL);
    _makepath(szSearchDir, szDrive, szDir, NULL, NULL);
    
    // Loads the DbgHelp.dll and retrieve pointers to functions
    // In MSVC 5.0, we have to get pointers to function dynamically, because the
    // newest versions of DbgHelp.lib is not compilable with VC++ 5.0
    if((result = LoadDbgHelp()) != ERROR_SUCCESS)
    {
        strcpy(szFileName, "DbgHelp.dll not found");
        return result;
    }

    // Load and initialize symbols for the current process
    // We have to use the directory where the debug info file (.PDB) is stored.
    // This is the "Debug" directory by default.
    if(_SymInitialize(hProcess, szSearchDir, TRUE) == FALSE)
    {
        strcpy(szFileName, "Cannot load program symbols");
        return GetLastError();
    }

    // Set the options for symbol library, otherwise the _SymGetLineFromAddr will not work
    _SymSetOptions(SYMOPT_LOAD_LINES);

    // Fill-in required members of IMAGEHLP_LINE
    ZeroMemory(&il, sizeof(IMAGEHLP_LINE));
    il.SizeOfStruct = sizeof(IMAGEHLP_LINE);
    bResult = _SymGetLineFromAddr(hProcess, (DWORD)pAddr, &dwDisplacement, &il);

    // Try to retrieve source file name and the line number.
    if(bResult)
    {
        // WOW, we succeeded !!!
        dwLine = il.LineNumber;
        strcpy(szFileName, il.FileName);
        result = ERROR_SUCCESS;
    }
    else
    {
        result = GetLastError();
        strcpy(szFileName, "Not the debug version");
    }

    // Unload the module information and free the DLL
    _SymCleanup(hProcess);
    UnloadDbgHelp();
    return result;
}

// This function will help us to detect the place of an exception
// It will show the detailed exception information (Exception code, address,
// source file name and line number). If the exception is continuable,
// the user will be asked if he wants to continue program execution.
int ShowExceptionInfo(EXCEPTION_POINTERS * ep)
{
    char  szMessage[256];
    char  szModule[MAX_PATH];           // Module where the exception occured
    char  szFileName[MAX_PATH];         // Source file name
    char  temp[MAX_PATH];
    DWORD dwLine;                       // Line where the exception occured
    UINT  flags = MB_OK+MB_ICONERROR;
    int   iYesResult = EXCEPTION_CONTINUE_EXECUTION;
    int   result;

    // Retrieve the module name where the exception was raised
    if(GetModuleFromAddr(ep->ExceptionRecord->ExceptionAddress, szModule) != NULL)
    {
        strcpy(temp, strrchr(szModule, '\\') + 1);
        strcpy(szModule, temp);
    }
    else
        strcpy(szModule, "(unknown)");
    
    // Exception code and the address is available in both debug and release version.
    sprintf(szMessage, "%s (0x%08lX) in module %s, address %p.\n",
                       GetExceptionName(ep->ExceptionRecord->ExceptionCode),
                       ep->ExceptionRecord->ExceptionCode,
                       szModule,
                       ep->ExceptionRecord->ExceptionAddress);

    // If source file name and line number are available
    if((result = GetSourceNameAndLineNum(ep->ExceptionRecord->ExceptionAddress, szFileName, dwLine)) == ERROR_SUCCESS)
    {
        sprintf(temp, "File : %s, Line : %u\n", szFileName, dwLine);
        strcat(szMessage, temp);
    }
    else
    {
        sprintf(temp, "(Cannot get file name and line no.: %s)\n", szFileName);
        strcat(szMessage, temp);
    }

#ifdef _MSC_VER
    // If we run under a debugger, we allow the programmer to re-raise the exception into
    // the debugger. This come handy when the programmer needs to go directly to the
    // the exception occurence place.
    if(TestDebugger())
    {
        flags = MB_YESNO+MB_ICONERROR;
        strcat(szMessage, "Do you wish to re-raise the exception into the current debugger ?");
        iYesResult = EXCEPTION_CONTINUE_SEARCH;
    }
    else
#endif

    // Test if we can continue after an exception
    if(ep->ExceptionRecord->ExceptionFlags == 0)
    {
        strcat(szMessage, "Do you wish to retry the execution of code that caused the exception ?");
        flags = MB_YESNO+MB_ICONERROR;
    }

    // Show the exception information
    if(MessageBoxA(NULL, szMessage, "Program caused an exception", flags) == IDYES)
        return iYesResult;

    return EXCEPTION_EXECUTE_HANDLER;
}

// If the __TEST_PROGRAM__ is defined, this module will act like a testprogram with its own
// main() function. If you want to use this module as a part of another project, don't define this
// constant
#ifdef __TEST_PROGRAM__
void main(void)
{
    EXCEPTION_POINTERS * ep;

    // Start the frame for an exception.
    __try
    {
//      FARPROC BadFunction = NULL;
//      DWORD   x = 1, y = 0;

        // Division by zero
//      x = x / y;

        // Access violation
        *((DWORD *)0) = 0;

        // Call of null-pointer function
//      BadFunction();
    }

    // Exception handler
    __except(ep = GetExceptionInformation(), ShowExceptionInfo(ep))
    {}
}
#endif // __TEST_PROGRAM__
